#   Copyright (C) 2020  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>


import cv2
import cv2.aruco as aruco
import numpy as np
import dlib
import os
import sys
import logging
import tobiiglasses
from collections import OrderedDict
from tobiiglasses.aoi.model import AOI, AOI_Item


class Aruco_Detector:

    def __init__(self, cameraMatrix, distCoeffs):
        self.__cameraMatrix__ = cameraMatrix
        self.__distCoeffs__ = distCoeffs

    def getCorners(self, opencvMat, auro_dict):
        opencvMat = cv2.cvtColor(opencvMat, cv2.COLOR_BGR2GRAY)
        corners, ids, rejectedImgPoints = aruco.detectMarkers(opencvMat, auro_dict)
        return (corners, ids)

    def getDetectedFeatures(self, board, corners, ids):
        (valid, rvec, tvec) = board.estimatePose(corners, ids, self.__cameraMatrix__, self.__distCoeffs__)
        detected_features_points = None
        if valid > 0:
            imgpts, jac = cv2.projectPoints(board.feature_3dpoints, rvec, tvec, self.__cameraMatrix__, self.__distCoeffs__)
            detected_features_points = np.array([[int(imgpts[0][0][0]), int(imgpts[0][0][1])],
                                                 [int(imgpts[1][0][0]), int(imgpts[1][0][1])],
                                                 [int(imgpts[2][0][0]), int(imgpts[2][0][1])],
                                                 [int(imgpts[3][0][0]), int(imgpts[3][0][1])]  ])
        return detected_features_points


class AOI_Aruco:
    MARKERS_X = 3#7
    MARKERS_Y = 2#4
    ARUCO_DICT = aruco.Dictionary_get(aruco.DICT_4X4_100)

    def __init__(self, aoi_label, aoi_id, aruco_detector, markerLength=0.1, markerSeparation=0.1, width=1920, height=1080):
        self.aoi_label = aoi_label
        self.aoi_id = aoi_id
        self.markerLength = markerLength
        self.markerSeparation = markerSeparation
        self.__width__ = width
        self.__height__ = height

        self.feature_3dpoints = np.array( [[self.markerLength, AOI_Aruco.MARKERS_Y*self.markerLength + (AOI_Aruco.MARKERS_Y-1)*self.markerSeparation, 0.0],
                                           [(AOI_Aruco.MARKERS_X-1)*self.markerLength + (AOI_Aruco.MARKERS_X-1)*self.markerSeparation, AOI_Aruco.MARKERS_Y*self.markerLength + (AOI_Aruco.MARKERS_Y-1)*self.markerSeparation, 0.0],
                                           [self.markerLength, 0.0, 0.0],
                                           [(AOI_Aruco.MARKERS_X-1)*self.markerLength + (AOI_Aruco.MARKERS_X-1)*self.markerSeparation, 0.0, 0.0] ])

        self.__aoi__ = aruco.GridBoard_create(markersX=AOI_Aruco.MARKERS_X,
                                   markersY=AOI_Aruco.MARKERS_Y,
                                   markerLength=self.markerLength,
                                   markerSeparation=self.markerSeparation,
                                   dictionary=AOI_Aruco.ARUCO_DICT,
                                   firstMarker=aoi_id*AOI_Aruco.MARKERS_X*AOI_Aruco.MARKERS_Y)

        corners, self.__ids__ = aruco_detector.getCorners(self.getCVMat(), AOI_Aruco.ARUCO_DICT)
        self.features_2dpoints = aruco_detector.getDetectedFeatures(self, corners, self.__ids__)

    def estimatePose(self, corners, ids, cameraMatrix, distCoeffs):
        rvec = np.array( [0.0, 0.0, 0.0] )
        tvec = np.array( [0.0, 0.0, 0.0] )
        (valid, rvec, tvec) = aruco.estimatePoseBoard(corners, ids, self.__aoi__, cameraMatrix, distCoeffs, rvec, tvec)
        return (valid, rvec, tvec)

    def exportAOI(self, filepath):
        img = self.getCVMat()
        cv2.rectangle(img,(self.features_2dpoints[0][0],self.features_2dpoints[0][1]),(self.features_2dpoints[3][0],self.features_2dpoints[3][1]),(255,255,255),-1)
        cv2.imwrite(os.path.join(filepath, self.getAOIFilename()), img)

    def getAOI(self):
        return self.__aoi__

    def getCVMat(self):
        img = np.zeros([self.__height__, self.__width__, 1],dtype=np.uint8)
        img.fill(255)
        self.__aoi__.draw((self.__width__, self.__height__), img, 40)
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img

    def getAOIFilename(self):
        return "aruco_%s.png" % self.aoi_label

class AOI_Aruco_Model(AOI):
    def __init__(self, cameraMatrix, distCoeffs):
        AOI.__init__(self)
        self.__cameraMatrix__ = cameraMatrix
        self.__distCoeffs__ = distCoeffs
        self.__aoi_boards__ = {}
        self.__aruco_detector__ = Aruco_Detector(cameraMatrix, distCoeffs)


    def getDetectedItems(self, opencvMat):
        (corners, ids) = self.__aruco_detector__.getCorners(opencvMat, AOI_Aruco.ARUCO_DICT)
        AOI_Items = []
        if ids is None:
            return AOI_Items
        for label, board in self.__aoi_boards__.items():
            filtered_corners = []
            filtered_ids = []
            for marker_id in ids:
                if marker_id in board.__ids__:
                    filtered_corners.append(corners[ids.tolist().index(marker_id)])
                    filtered_ids.append(marker_id)
            filtered_ids = np.asarray(filtered_ids, dtype=np.int32)
            filtered_corners = np.asarray(filtered_corners, dtype=np.float32)
            if len(filtered_ids) > 0:
                detected_features_points = self.__aruco_detector__.getDetectedFeatures(board, filtered_corners, filtered_ids)
                if detected_features_points is None:
                    break
                else:
                    item = AOI_Item(label, detected_features_points, board.getAOIFilename(), board.features_2dpoints)
                    AOI_Items.append(item)

        return AOI_Items

    def createArucoAOI(self, aoi_label, markerLength=0.1, markerSeparation=0.1, width=1920, height=1080):
        if aoi_label in self.__aoi_boards__.keys():
            logging.error('aoi_label is already present. AOI will not be created!')
        else:
            self.__aoi_boards__[aoi_label] = AOI_Aruco(aoi_label, len(self.__aoi_boards__.keys()), self.__aruco_detector__, markerLength, markerSeparation, width, height)
            self.__aoi_boards__[aoi_label].exportAOI('./')

    def exportArucoAOIs(self, filepath='./'):
        for label, board in self.__aoi_boards__.items():
            board.exportAOI(filepath)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import numpy as np
import cv2
import matplotlib.path as mplPath
from tobiiglasses.aoi.heatmaps import Heatmap

class AOI_Item:

    def __init__(self, aoi_id, detected_features_points, snapshot_filename, features_points,landmarks=[], aoi_score=100):
        self.aoi_id = aoi_id
        self.detected_features_points = detected_features_points
        self.snapshot_filename = snapshot_filename
        self.snapshot = cv2.imread(snapshot_filename)
        self.features_points = features_points
        self.landmarks = landmarks
        self.aoi_score = aoi_score


class AOI:

    def __init__(self):
        self.__aoi_regions__ = {}
        self.__aoi_hits__ = {}
        self.__aoi_items__ = {}

    def apply(self, opencvMat, ts, gaze_x, gaze_y, gaze_events):
        for item in self.getDetectedItems(opencvMat):
            detected_features_points = item.detected_features_points
            aoi_id = item.aoi_id
            if not aoi_id in self.__aoi_items__.keys():
                self.__aoi_items__[aoi_id] = item
            if not aoi_id in self.__aoi_regions__.keys():
                self.__aoi_regions__[aoi_id] = {}
                self.__aoi_hits__[aoi_id] = {}

            H, status = cv2.findHomography(item.features_points, detected_features_points)
            rows, cols, ch = item.snapshot.shape

            #bounding_box_src = np.array([[0,0,1], [0, rows-1,1], [cols-1, rows-1,1], [cols-1, 0,1]])
            bounding_box_src = np.array([[item.features_points[0][0],item.features_points[0][1],1],
                                        [item.features_points[1][0], item.features_points[1][1],1],
                                        [item.features_points[3][0], item.features_points[3][1],1],
                                        [item.features_points[2][0], item.features_points[2][1],1]])
            bounding_box_dst = H.dot(bounding_box_src.T)

            bounding_box_dst = np.array([[bounding_box_dst[0,0]/bounding_box_dst[2,0], bounding_box_dst[1,0]/bounding_box_dst[2,0]],
                                             [bounding_box_dst[0,1]/bounding_box_dst[2,1], bounding_box_dst[1,1]/bounding_box_dst[2,1]],
                                             [bounding_box_dst[0,2]/bounding_box_dst[2,2], bounding_box_dst[1,2]/bounding_box_dst[2,2]],
                                             [bounding_box_dst[0,3]/bounding_box_dst[2,3], bounding_box_dst[1,3]/bounding_box_dst[2,3]] ])

            self.__aoi_regions__[aoi_id][ts] = bounding_box_dst

            if self.contains(ts, aoi_id, gaze_x, gaze_y):
                p = [gaze_x, gaze_y, 1]
                q = np.linalg.inv(H).dot(p)
                q /= q[2]
                self.__aoi_hits__[aoi_id][ts] = [q[0], q[1]]
                gaze_events.setAOI(ts, int(q[0]), int(q[1]), aoi_id, item.aoi_score)
            else:
                gaze_events.setAOI(ts, 0, 0, aoi_id, 0)

            self.showLandmarks(opencvMat, item.landmarks)

    def getDetectedItems(self, opencvMat):
        raise NotImplementedError( "AOIs should have implemented getDetectedItems() method" )

    def contains(self, ts, aoi_id, gaze_x, gaze_y, threshold=-60.0):
        res = False
        if ts in self.__aoi_regions__[aoi_id].keys():
            res = mplPath.Path(self.__aoi_regions__[aoi_id][ts]).contains_point((gaze_x, gaze_y), radius=threshold)
        return res

    def drawAOIsBox(self, opencvMat, ts, color=(255, 0, 0)):
        for aoi_id in self.__aoi_regions__.keys():
            if ts in self.__aoi_regions__[aoi_id].keys():
                aoi_points = self.__aoi_regions__[aoi_id][ts]
                cv2.polylines(opencvMat, np.int32([aoi_points]), 1, color)

    def getAOIRegions(self, ts, aoi_id):
        if ts in self.__aoi_regions__[aoi_id].keys():
            return self.__aoi_regions__[aoi_id][ts]
        return None

    def getAOIHits(self, aoi_id):
        return self.__aoi_hits__[aoi_id]

    def getAOI(self, ts, gaze_x, gaze_y):
        for aoi_id in self.__aoi_regions__:
            if self.contains(ts, aoi_id, gaze_x, gaze_y) is True:
                return aoi_id
        return None

    def exportHeatmap(self):
        H = Heatmap()
        for aoi_id, aoi_item in self.__aoi_items__.items():
            rows, cols, ch = aoi_item.snapshot.shape
            dispsize = (int(cols), int(rows))
            H.draw( list(self.__aoi_hits__[aoi_id].values()), dispsize, aoi_item.snapshot_filename, savefilename=('%s_%s.png' % ('heatmap', aoi_id)) )

    def showLandmarks(self, opencvMat, landmarks, color=(0, 0, 255)):
        for (x,y) in landmarks:
            cv2.circle(opencvMat, (x, y), 2, (0, 0, 255), -1)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

from tobiiglasses.gazedata import GazeData

class VideoAndGaze:

    def __init__(self, opencvMat, x, y, ts):
        self.opencvMat = opencvMat
        self.x = x
        self.y = y
        self.ts = ts

class VideoFramesAndMappedGaze:

    def __init__(self, gazedata, videoCapture, fps):
        self.__cap__ = videoCapture
        if (self.__cap__.isOpened() == False):
            print("Error opening video stream or file")
            raise IOError
        self.__current_frameAndMappedGaze__ = None
        self.__df__ = gazedata.toDataFrame()
        self.__vts__ = gazedata.getVTS()
        self.__vts_list__ = self.__vts__.keys()
        self.__frame_duration__ = int(1000/fps)
        self.__current_time__ = self.__vts_list__[0]
        self.__i__ = 0

    def __iter__(self):
        return self

    def __next__(self):
        if (self.__cap__.isOpened()):
            ret, frame = self.__cap__.read()
            if ret == True:
                if self.__current_time__ > self.__vts_list__[self.__i__]:
                    if self.__i__ < len(self.__vts_list__) - 1:
                        self.__i__+=1
                delay = int(self.__vts__[self.__vts_list__[self.__i__]])
                T = self.__df__.index[self.__df__.index.get_loc(self.__current_time__+self.__i__, method='nearest')]
                x = self.__df__.at[T, GazeData.GazePixelX]
                y = self.__df__.at[T, GazeData.GazePixelY]
            else:
                raise StopIteration
            self.__current_time__ += self.__frame_duration__
            return (frame, x, y, T)
        else:
            raise StopIteration



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import logging
import os
import csv
import gzip
import numpy as np
import pandas as pd
import cv2
import tobiiglasses.entities
from tobiiglasses.entities import TobiiRecording, TobiiSegment
from tobiiglasses.gazedata import GazeData
from tobiiglasses.events import GazeEvents
from tobiiglasses.exporter import RawCSV, ExtendedRawCSV
from tobiiglasses.video import VideoFramesAndMappedGaze, VideoAndGaze
from tobiiglasses.metrics import Fixations_Metrics

import cv2.aruco as aruco

class Recording:

    def __init__(self, root_dir, project_id, recording_id):
        recordings_path = tobiiglasses.entities.get_recordings_path(root_dir, project_id)
        self.__recording__ = TobiiRecording(recordings_path, recording_id)
        self.__segment_data__ = {}
        self.__segment_events__ = {}
        self.__segment_ids__ = []
        self.__loadSegmentIDs__()

    def __exportData_CSV__(self, filepath=None, filename=None, csv_suffix='', segment_id=None, exporter=None):
        segment_ids = self.__getSegmentIDs__(segment_id)
        for s_id in segment_ids:
            gazedata = self.getGazeData(s_id)
            (filepath, filename) = self.__getFileParams__('csv', s_id, filepath, filename, csv_suffix)
            e = exporter(filepath, filename, gazedata)
            e.toCSV()

    def __getFileParams__(self, extension, segment_id=None, filepath=None, filename=None, suffix=''):
        if segment_id is None:
            segment_id = 1
        if filepath is None:
            filepath = "."
        if filename is None:
            filename = "%s-S%s-%s.%s" % (self.getParticipantName(), str(segment_id), suffix, extension)
        return (filepath, filename)

    def __getSegmentIDs__(self, segment_id=None):
        segment_ids = []
        if segment_id is None:
            segment_ids.extend(self.__segment_ids__)
        else:
            segment_ids.append(segment_id)
        return segment_ids

    def __loadSegmentIDs__(self):
        self.__segment_ids__.extend(range(1, self.__recording__.getSegmentsN() + 1))

    def exportFull(self, fixation_filter, filepath=None, csv_filename='output.csv', video_filename='output.avi', segment_id=1, aoi_models=[]):
        fixations = self.getFixations(fixation_filter, ts_filter=None, segment_id=segment_id)
        if filepath is None:
            filepath = "."
        logging.info('Exporting video with mapped fixations in folder %s' % filepath)
        data = self.getGazeData(segment_id)
        fps = data.getFrameFPS()
        width = data.getFrameWidth()
        height = data.getFrameHeight()
        f = self.__recording__.getSegment(segment_id).getVideoFilename()
        cap = cv2.VideoCapture(f)
        framesAndGaze = iter(VideoFramesAndMappedGaze(data, cap, fps))
        fourcc = cv2.VideoWriter_fourcc(*'XVID')
        out = cv2.VideoWriter(video_filename,fourcc, fps, (width,height))


        for frame, x, y, ts in framesAndGaze:
            if not (x > 0 or y > 0):
                continue
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            aruco_dict = aruco.Dictionary_get(aruco.DICT_4X4_50)
            parameters =  aruco.DetectorParameters_create()
            corners, ids, rejectedImgPoints = aruco.detectMarkers(gray, aruco_dict, parameters=parameters)
            (fx, fy, duration) = fixations.getClosestFixation(ts)

            for model in aoi_models:
                model.apply(frame, ts, fx, fy, fixations)
                model.drawAOIsBox(frame, ts)

            if fx>0 and fy>0:
                color = (0,0,255)
                for model in aoi_models:
                    aoi_id = model.getAOI(ts, fx, fy)
                    if not aoi_id is None:
                        color = (0,255,0)
                        cv2.putText(frame, aoi_id, (fx, fy), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 255, 0), 2)
                cv2.circle(frame,(fx,fy), 30, color , 2)
            frame_markers = aruco.drawDetectedMarkers(frame.copy(), corners, ids)
            out.write(frame_markers)
        cap.release()
        model.exportHeatmap()
        (filepath, filename) = self.__getFileParams__('csv', segment_id, filepath, csv_filename, 'Fixations')
        fixations.exportCSV(filepath, filename, ts_filter=None)
        f_metrics = Fixations_Metrics(fixations)


    def exportCSV_ExtendedRawData(self, filepath=None, filename=None, segment_id=None):
        self.__exportData_CSV__(filepath, filename, 'eRaw', segment_id, ExtendedRawCSV)

    def exportCSV_Fixations(self, fixation_filter, filepath=None, filename=None, ts_filter=None, segment_id=None):
        logging.info('Exporting fixations on %s' % filepath)
        fixations = self.getFixations(fixation_filter, ts_filter, segment_id)
        (filepath, filename) = self.__getFileParams__('csv', segment_id, filepath, filename, 'Fixations')
        fixations.exportCSV(filepath, filename, ts_filter=ts_filter)
        return fixations

    def exportCSV_RawData(self, filepath=None, filename=None, segment_id=1):
        self.__exportData_CSV__(filepath, filename, 'Raw', segment_id, RawCSV)

    def exportVideoAndGaze(self, filepath=None, filename='output.avi', segment_id=1, aoi_dnn_models=[]):
        logging.info('Exporting video with mapped gaze in file %s in folder %s' % (filename, filepath))
        data = self.getGazeData(segment_id)
        fps = data.getFrameFPS()
        width = data.getFrameWidth()
        height = data.getFrameHeight()
        f = self.__recording__.getSegment(segment_id).getVideoFilename()
        cap = cv2.VideoCapture(f)
        framesAndGaze = iter(VideoFramesAndMappedGaze(data, cap, fps))
        fourcc = cv2.VideoWriter_fourcc(*'XVID')
        out = cv2.VideoWriter(filename,fourcc, fps, (width,height))
        for frame, x, y, ts in framesAndGaze:
            for model in aoi_dnn_models:
                model.apply(frame, ts, x, y)
                model.drawAOIsBox(frame, ts)

            if x>0 and y>0:
                color = (0,0,255)
                aoi_id = model.getAOI(ts, x, y)
                if not aoi_id is None:
                    color = (0,255,0)
                    cv2.putText(frame, aoi_id, (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 255, 0), 2)
                cv2.circle(frame,(x,y), 30, color , 2)
            out.write(frame)
        cap.release()
        model.exportHeatmap()

    def getFixations(self, fixation_filter, ts_filter=None, segment_id=None):
        segment_ids = self.__getSegmentIDs__(segment_id)
        for s_id in segment_ids:
            gaze_data = self.getGazeData(s_id)
            gaze_events = self.getGazeEvents(s_id)
            gaze_events.filterFixations(fixation_filter, gaze_data.toDataFrame(), ts_filter)
        return gaze_events

    def getGazeData(self, segment_id=1):
        segment = self.__recording__.getSegment(segment_id)
        if not segment_id in self.__segment_data__.keys():
            self.__segment_data__[segment_id] = GazeData(segment)
        return self.__segment_data__[segment_id]

    def getGazeEvents(self, segment_id):
        segment = self.__recording__.getSegment(segment_id)
        if not segment_id in self.__segment_events__.keys():
            self.__segment_events__[segment_id] = GazeEvents()
            events = self.getGazeData(segment_id).getLoggedEvents()
            for ts, value in events.items():
                self.__segment_events__[segment_id].addLoggedEvent(ts, value)
        return self.__segment_events__[segment_id]

    def getParticipantName(self):
        return self.__recording__.getParticipant().getName()

    def replay(self, segment_id):
        logging.info('Replaying video with mapped gaze...')
        data = self.getGazeData(segment_id)
        fps = data.getFrameFPS()
        width = data.getFrameWidth()
        height = data.getFrameHeight()
        f = self.__recording__.getSegment(segment_id).getVideoFilename()
        cap = cv2.VideoCapture(f)
        framesAndGaze = iter(VideoFramesAndMappedGaze(data, cap, fps))
        for frame, x, y, ts in framesAndGaze:
            if x>0 and y>0:
                cv2.circle(frame,(int(x),int(y)), 60, (0,0,255), 2)
            cv2.imshow('Frame',frame)
            if cv2.waitKey(25) & 0xFF == ord('q'):
                break
        cap.release()
        cv2.destroyAllWindows()

    def saveVideoSnapshot(self, filename, ts, segment_id):
        logging.info('Saving snapshot from video recording in %s ' % filename)
        segment = self.__recording__.getSegment(segment_id)
        fps = segment.getFrameFPS()
        f = segment.getVideoFilename()
        cap = cv2.VideoCapture(f)
        framesAndGaze = iter(VideoFramesAndMappedGaze(data, cap, fps))
        for frame, x, y, _ts in framesAndGaze:
            if _ts > ts:
                cv2.imwrite(filename,frame)
                break
        cap.release()



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

from model import AOI_Model

class AOI_SURFModel(AOI_Model):

    def __init__(self):
        pass

    def fit(self, gaze_events, ts_filter):
        pass



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>


import cv2
import numpy as np
import dlib
import os
import sys
import tobiiglasses
from collections import OrderedDict
#from tobiiglasses.aoi.dnn.model import AOI_DNN_Model
from tobiiglasses.aoi.model import AOI, AOI_Item

FACIAL_LANDMARKS_IDXS = OrderedDict([
        ("mouth", (48, 67)),
        ("right_eyebrow", (17, 21)),
        ("left_eyebrow", (22, 26)),
        ("right_eye", (36, 41)),
        ("left_eye", (42, 47)),
        ("nose", (27, 35)),
        ("jaw", (0, 16))
        ])

FILENAME_SHAPE_PREDICTOR = os.path.join(tobiiglasses.aoi.dnn.__path__[0], 'faces', 'shape_predictor_68_face_landmarks.dat')
FILENAME_FACE_LANDMARKS = os.path.join(tobiiglasses.aoi.dnn.__path__[0], 'faces', 'facial_landmarks_68markup.png')
FACE_FEATURES_POINTS = np.array( [[132, 60], [330, 620], [515, 615], [675, 50]]) #ids = 17, 3, 13, 26

class Face:

    def __init__(self, left, top, right, bottom, face_landmarks):
        self.__left__ = left
        self.__top__ = top
        self.__right__ = right
        self.__bottom__ = bottom
        self.__landmarks__ = face_landmarks

        self.__mouth__ = face_landmarks['mouth']
        self.__right_eyebrow__ = face_landmarks['right_eyebrow']
        self.__left_eyebrow__ = face_landmarks['left_eyebrow']
        self.__right_eye__ = face_landmarks['right_eye']
        self.__left_eye__ = face_landmarks['left_eye']
        self.__nose__ = face_landmarks['nose']
        self.__jaw__ = face_landmarks['jaw']

    @property
    def top(self):
        return self.__top__

    @property
    def right(self):
        return self.__right__

    @property
    def left(self):
        return self.__left__

    @property
    def bottom(self):
        return self.__bottom__

    @property
    def landmarks(self):
        return self.__landmarks__

    @property
    def mouth(self):
        return self.__mouth__

    @property
    def right_eyebrow(self):
        return self.__right_eyebrow__

    @property
    def left_eyebrow(self):
        return self.__left_eyebrow__

    @property
    def right_eye(self):
        return self.__right_eye__

    @property
    def left_eye(self):
        return self.__left_eye__

    @property
    def jaw(self):
        return self.__jaw__

    @property
    def nose(self):
        return self.__nose__


class FaceLandmarksDetector:

    def __init__(self, filename_shape_predictor):
        self.__detector__ = dlib.get_frontal_face_detector()
        self.__predictor__ = dlib.shape_predictor(filename_shape_predictor)

    def getFaces(self, opencvMat):
        gray = cv2.cvtColor(opencvMat, cv2.COLOR_BGR2GRAY)
        faces = self.__detector__(gray)
        Faces = []
        for face in faces:
            landmarks = self.__predictor__(gray, face)
            FACE_LANDMARKS_XY = {}
            for (name, (i, j)) in FACIAL_LANDMARKS_IDXS.items():
                FACE_LANDMARKS_XY[name] = []
                for k in range(i, j+1):
                    x = landmarks.part(k).x
                    y = landmarks.part(k).y
                    FACE_LANDMARKS_XY[name].append((x,y))
            Faces.append( Face(face.left(), face.top(), face.right(), face.bottom(), FACE_LANDMARKS_XY) )
        return Faces


class AOI_Face_Model(AOI):

    def __init__(self):
        AOI.__init__(self)
        self.__detector__ = FaceLandmarksDetector(FILENAME_SHAPE_PREDICTOR)

    def getDetectedItems(self, opencvMat):
        faces = self.__detector__.getFaces(opencvMat)
        AOI_Items = []
        i=0
        for face in faces:
            i+=1
            detected_features_points = np.array([[face.landmarks['right_eyebrow'][0][0], face.landmarks['right_eyebrow'][0][1]],
                                                 [face.landmarks['jaw'][7][0], face.landmarks['jaw'][7][1]],
                                                 [face.landmarks['jaw'][9][0], face.landmarks['jaw'][9][1]],
                                                 [face.landmarks['left_eyebrow'][4][0], face.landmarks['left_eyebrow'][4][1]]
                                                ])
            """
            landmarks = []
            for name in face.landmarks.keys():
                for p in face.landmarks[name]:
                    landmarks.append( [p[0], p[1]])
            """
            AOI_Items.append( AOI_Item('face_' + str(i), detected_features_points, FILENAME_FACE_LANDMARKS, FACE_FEATURES_POINTS) )
        return AOI_Items



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# The following code is based on the PyGazeAnalyser/pygazeanalyser/gazeplotter.py
#   available at github.com/esdalmaijer/PyGazeAnalyser

import os
import argparse
import csv
import numpy
import matplotlib
from matplotlib import pyplot, image


class Heatmap:

    def draw_display(self, dispsize, imagefile=None):
        # construct screen (black background)
        screen = numpy.zeros((dispsize[1], dispsize[0], 3), dtype='float32')
        # if an image location has been passed, draw the image
        if imagefile != None:
            # check if the path to the image exists
            if not os.path.isfile(imagefile):
                raise Exception("ERROR in draw_display: imagefile not found at '%s'" % imagefile)
            # load image
            img = image.imread(imagefile)

            # width and height of the image
            w, h = len(img[0]), len(img)
            # x and y position of the image on the display
            x = int(dispsize[0] / 2 - w / 2)
            y = int(dispsize[1] / 2 - h / 2)
            # draw the image on the screen
            screen[y:y + h, x:x + w, :] += img
        # dots per inch
        dpi = 100.0
        # determine the figure size in inches
        figsize = (dispsize[0] / dpi, dispsize[1] / dpi)
        # create a figure
        fig = pyplot.figure(figsize=figsize, dpi=dpi, frameon=False)
        ax = pyplot.Axes(fig, [0, 0, 1, 1])
        ax.set_axis_off()
        fig.add_axes(ax)
        # plot display
        ax.axis([0, dispsize[0], 0, dispsize[1]])
        ax.imshow(screen)  # , origin='upper')

        return fig, ax

    def gaussian(self, x, sx, y=None, sy=None):
        # square Gaussian if only x values are passed
        if y == None:
            y = x
        if sy == None:
            sy = sx
        # centers
        xo = x / 2
        yo = y / 2
        # matrix of zeros
        M = numpy.zeros([y, x], dtype=float)
        # gaussian matrix
        for i in range(x):
            for j in range(y):
                M[j, i] = numpy.exp(
                    -1.0 * (((float(i) - xo) ** 2 / (2 * sx * sx)) + ((float(j) - yo) ** 2 / (2 * sy * sy))))

        return M

    def draw(self, gazepoints, dispsize, imagefile=None, alpha=0.5, savefilename=None, gaussianwh=200, gaussiansd=None):
        # IMAGE
        fig, ax = self.draw_display(dispsize, imagefile=imagefile)

        # HEATMAP
        # Gaussian
        gwh = gaussianwh
        gsdwh = gwh / 6 if (gaussiansd is None) else gaussiansd
        gaus = self.gaussian(gwh, gsdwh)
        # matrix of zeroes
        strt = int(gwh / 2)
        heatmapsize = dispsize[1] + 2 * strt, dispsize[0] + 2 * strt
        heatmap = numpy.zeros(heatmapsize, dtype=float)
        # create heatmap
        for i in range(0, len(gazepoints)):
            # get x and y coordinates
            x = strt + int(gazepoints[i][0]) - int(gwh / 2)
            y = strt + int(gazepoints[i][1]) - int(gwh / 2)
            # correct Gaussian size if either coordinate falls outside of
            # display boundaries
            if (not 0 < x < dispsize[0]) or (not 0 < y < dispsize[1]):
                hadj = [0, gwh];
                vadj = [0, gwh]
                if 0 > x:
                    hadj[0] = abs(x)
                    x = 0
                elif dispsize[0] < x:
                    hadj[1] = gwh - int(x - dispsize[0])
                if 0 > y:
                    vadj[0] = abs(y)
                    y = 0
                elif dispsize[1] < y:
                    vadj[1] = gwh - int(y - dispsize[1])
                # add adjusted Gaussian to the current heatmap
                try:
                    heatmap[y:y + vadj[1], x:x + hadj[1]] += gaus[vadj[0]:vadj[1], hadj[0]:hadj[1]] * 1
                except:
                    # fixation was probably outside of display
                    pass
            else:
                # add Gaussian to the current heatmap
                heatmap[y:y + gwh, x:x + gwh] += gaus * 1
        # resize heatmap
        heatmap = heatmap[strt:dispsize[1] + strt, strt:dispsize[0] + strt]
        # remove zeros
        lowbound = numpy.mean(heatmap[heatmap > 0])
        heatmap[heatmap < lowbound] = numpy.NaN
        # draw heatmap on top of image
        ax.imshow(heatmap, cmap='jet', alpha=alpha)

        # FINISH PLOT
        # invert the y axis, as (0,0) is top left on a display
        ax.invert_yaxis()
        # save the figure if a file name was provided
        if savefilename != None:
            fig.savefig(savefilename)

        return fig
